# -*- coding: utf-8 -*-
# 版权所有 2019 深圳米筐科技有限公司（下称“米筐科技”）
#
# 除非遵守当前许可，否则不得使用本软件。
#
#     * 非商业用途（非商业用途指个人出于非商业目的使用本软件，或者高校、研究所等非营利机构出于教育、科研等目的使用本软件）：
#         遵守 Apache License 2.0（下称“Apache 2.0 许可”），您可以在以下位置获得 Apache 2.0 许可的副本：http://www.apache.org/licenses/LICENSE-2.0。
#         除非法律有要求或以书面形式达成协议，否则本软件分发时需保持当前许可“原样”不变，且不得附加任何条件。
#
#     * 商业用途（商业用途指个人出于任何商业目的使用本软件，或者法人或其他组织出于任何目的使用本软件）：
#         未经米筐科技授权，任何个人不得出于任何商业目的使用本软件（包括但不限于向第三方提供、销售、出租、出借、转让本软件、本软件的衍生产品、引用或借鉴了本软件功能或源代码的产品或服务），任何法人或其他组织不得出于任何目的使用本软件，否则米筐科技有权追究相应的知识产权侵权责任。
#         在此前提下，对本软件的使用同样需要遵守 Apache 2.0 许可，Apache 2.0 许可与本许可冲突之处，以本许可为准。
#         详细的授权流程，请联系 public@ricequant.com 获取。

import simplejson as json
import datetime

from rqalpha import const


def convert_dict_to_json(dict_obj):
    dict_obj = json.dumps(dict_obj, default=custom_encode)
    return dict_obj


def convert_json_to_dict(json_str):
    dict_obj = json.loads(json_str, object_hook=custom_decode)
    return dict_obj


def custom_encode(obj):
    if isinstance(obj, datetime.datetime):
        obj = {"__datetime__": True, "as_str": obj.strftime("%Y%m%dT%H:%M:%S.%f")}
    elif isinstance(obj, datetime.date):
        obj = {"__date__": True, "as_str": obj.strftime("%Y%m%d")}
    elif isinstance(obj, const.CustomEnum):
        obj = {"__enum__": True, "as_str": str(obj)}
    else:
        raise TypeError(
            "Unserializable object {} of type {}".format(obj, type(obj))
        )
    return obj


def custom_decode(obj):
    if "__datetime__" in obj:
        obj = datetime.datetime.strptime(obj["as_str"], "%Y%m%dT%H:%M:%S.%f")
    elif "__date__" in obj:
        obj = datetime.datetime.strptime(obj["as_str"], "%Y%m%d").date()
    elif "__enum__" in obj:
        [e, v] = obj["as_str"].split(".")
        obj = getattr(getattr(const, e), v)
    return obj
